///////////////////////////////////////////////////////////////////////////////
// MUSICAL AGENT - Honours Thesis Project of Joseph Rukshan Fonseka
//                 Monash University Australia 2000
// Contact         jfon2@student.monash.edu.au
// Acknowledgments
//                 Dr. Alan Dorin
//                 Mr. I Kaminskyj
//                 Mr. Jon McCormack
//
// ~MIDI.java~
///////////////////////////////////////////////////////////////////////////////

import java.io.File;
import java.io.IOException;

import javax.sound.midi.Instrument;
import javax.sound.midi.MidiChannel;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Sequence;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Soundbank;
import javax.sound.midi.Synthesizer;
import javax.sound.midi.Track;
import javax.swing.JOptionPane;

/**
 * This provides simple MIDI functionality such as playing a note according to
 * specific parameters such as velocity, duration, intensity etc. Handling of
 * instruments is done automatically. If a new instrument is required, it is
 * assigned to a new channel. If all channels are filled up and a new instrument
 * is required, then that note message is assigned to the last channel. If a
 * midi synthesizer is not available, these midi routines have no effect. The
 * method open must be called before any midi operation. The method close must
 * be called to properly shut down the midi synthesizer. Also implemented in
 * this class is functionality to capture MIDI events such as noteon, noteoff
 * and programchanges. These events can then be saved to a MIDI file.
 */
public class MIDI {
	private Synthesizer synthesizer;
	private Instrument instruments[];
	private ChannelData channels[];
	private boolean MIDIFound;
	private Sequence sequence;
	private Track track;
	private byte volume = 90;
	private long startTime;
	private boolean record = false;

	/**
	 * information for a particular MIDI channel, provides information about
	 * what instrument has been allocated for a particular channel
	 */
	class ChannelData {
		MidiChannel channel;
		int num, velocity, instrument, status;

		public ChannelData(MidiChannel channel, int num) {
			this.channel = channel; // assign a midi channel to this
			this.num = num; // channel number
			this.instrument = -1; // instrument that has been assigned to this
									// channel
			this.status = 0; // if 0 channel is free
		}
	}

	/** closes the midi synthesizer */
	public void close() {
		// stop all notes currently playing on all channels
		if (MIDIFound) {
			for (int i = 0; i < channels.length; i++) {
				channels[i].channel.allNotesOff();
			}
		}

		if (synthesizer != null) {
			synthesizer.close();
		}

		synthesizer = null;
		instruments = null;
		channels = null;
	}

	/** opens the midi synthesizer */
	public void open() {
		MIDIFound = true;
		try {
			if (synthesizer == null) {
				if ((synthesizer = MidiSystem.getSynthesizer()) == null) {
					JOptionPane.showMessageDialog(null,
							"getSynthesizer() failed!");
					return;
				}
			}
			synthesizer.open();
			sequence = new Sequence(Sequence.PPQ, 10);
		}
		// if this fails, set a flag so that other methods don't make any midi
		// calls
		catch (MidiUnavailableException e) {
			MIDIFound = false;
			JOptionPane.showMessageDialog(null,
					"MIDI not supported or is busy. Ignoring midi");
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null, e);
			System.exit(1);
		}

		Soundbank sb = synthesizer.getDefaultSoundbank();

		// load default instruments and assign the first instrument. By default
		// this is the
		// piano instrument
		if (sb != null) {
			instruments = synthesizer.getDefaultSoundbank().getInstruments();
			synthesizer.loadInstrument(instruments[0]);
		}

		// set up the 16 channels
		MidiChannel midiChannels[] = synthesizer.getChannels();
		channels = new ChannelData[midiChannels.length];

		for (int i = 0; i < channels.length; i++) {
			channels[i] = new ChannelData(midiChannels[i], i);
		}
	}

	/** this sets the global MIDI volume for the performance */
	public void setVolume(byte volume) {
		this.volume = volume;
	}

	/**
	 * this plays a note on an instrument. If the instrument does not exist in
	 * any channel, it will automatically be assigned to a channel. If all
	 * channels are full, this note will be played on the last channel.
	 */
	public void playNote(int note, int duration, int velocity, int instr) {
		int freechannel = -1;
		int i = -1;

		// check if midi was initialized successfully
		if (MIDIFound) {
			// check channels if this instrument exists or whether there is an
			// empty channel
			synchronized (channels) {
				while (i != channels.length - 1) {
					i++;
					// if status is zero it means channel is free
					if (channels[i].status == 0)
						freechannel = i;
					if (channels[i].instrument == instr) {
						freechannel = -1;
						break;
					}
				}

				// Instrument does not exists and a free channel was found so
				// assign this
				// instrument to it
				if (freechannel >= 0) {
					channels[freechannel].channel.programChange(instr);
					channels[freechannel].instrument = instr;

					// if recording, record that we did a program change
					if (record) {
						recordProgramChangeEvent(freechannel, instr);
					}
					i = freechannel;
				}

				// Start a new thread to play this note with the appropriate
				// volume
				Note n = new Note(note, duration, velocity, channels[i]);
				n.start();
			}
		}
	}

	public MIDI.Note createNote(int pitch, int duration, int velocity) {
		return new Note(pitch, duration, velocity, null);
	}
	
	/**
	 * this plays a note on an instrument. If the instrument does not exist in
	 * any channel, it will automatically be assigned to a channel. If all
	 * channels are full, this note will be played on the last channel.
	 */
	public void playNote(Note n) {
		int freechannel = -1;
		int i = -1;

		// check if midi was initialized successfully
		if (MIDIFound) {
			// check channels if this instrument exists or whether there is an
			// empty channel
			synchronized (channels) {
				while (i != channels.length - 1) {
					i++;
					// if status is zero it means channel is free
					if (channels[i].status == 0)
						freechannel = i;
					if (channels[i].instrument == n.instrument) {
						freechannel = -1;
						break;
					}
				}

				// Instrument does not exists and a free channel was found so
				// assign this
				// instrument to it
				if (freechannel >= 0) {
					channels[freechannel].channel.programChange(n.instrument);
					channels[freechannel].instrument = n.instrument;

					// if recording, record that we did a program change
					if (record) {
						recordProgramChangeEvent(freechannel, n.instrument);
					}
					i = freechannel;
				}

				// Start a new thread to play this note with the appropriate
				// volume
				n.channel = channels[i];
				n.start();
			}
		}
	}

	// this plays a note in a thread allowing multiple notes to be played
	// simultaneously.
	// This is quite effecient because java threads are light-weight and do not
	// create
	// a new workspace and so is very cpu and memory friendly
	class Note extends Thread {
		int note, duration, instrument, velocity;
		ChannelData channel;

		public Note(int note, int duration, int velocity, ChannelData channel) {
			this.note = note;
			this.duration = duration;
			this.velocity = velocity;
			this.channel = channel;
		}

		public void run() {
			// if recording, record the note on message
			if (record) {
				recordNoteEvent(ShortMessage.NOTE_ON, channel.num, note,
						velocity);
			}

			// status helps determine if the channel is free
			synchronized (channel) {
				channel.channel.noteOn(note, (volume * velocity) / 127);
				channel.status++;
				channel.velocity = velocity;
			}
			try {
				Thread.sleep(duration);
			} catch (InterruptedException e) {
			}

			// if recording, record the note off message
			if (record) {
				recordNoteEvent(ShortMessage.NOTE_OFF, channel.num, note,
						velocity);
			}
			synchronized (channel) {
				channel.channel.noteOff(note, (volume * velocity) / 127);
				channel.status--;
			}
		}
	}

	// this records note on and off messages in a track
	private void recordNoteEvent(int type, int channel, int note, int velocity) {
		ShortMessage message = new ShortMessage();
		try {
			long millis = System.currentTimeMillis() - startTime;
			long tick = millis * sequence.getResolution() / 500;
			message.setMessage(type + channel, note, velocity);
			MidiEvent event = new MidiEvent(message, tick);
			track.add(event);
		} catch (Exception ex) {
			JOptionPane.showMessageDialog(null, ex);
		}
	}

	// this records program changes in a track
	private void recordProgramChangeEvent(int channel, int instrument) {
		ShortMessage message = new ShortMessage();
		try {
			long millis = System.currentTimeMillis() - startTime;
			long tick = millis * sequence.getResolution() / 500;
			message.setMessage(ShortMessage.PROGRAM_CHANGE + channel,
					instrument, 0);
			MidiEvent event = new MidiEvent(message, tick);
			track.add(event);
		} catch (Exception ex) {
			JOptionPane.showMessageDialog(null, ex);
		}
	}

	/**
	 * this starts capturing the MIDI events, returns true if recording is in
	 * progress
	 */
	public boolean startRecording() {
		// get rid of previous recording if such exists
		if (track != null) {
			sequence.deleteTrack(track);
		}

		// make sure there is MIDI
		if (!MIDIFound) {
			JOptionPane.showMessageDialog(null, "No midi device!");
			return false;
		}
		record = true;
		track = sequence.createTrack();
		startTime = System.currentTimeMillis();
		return true;
	}

	/** this stops the recording */
	public void stopRecording() {
		// make sure there is MIDI
		if (!MIDIFound) {
			JOptionPane.showMessageDialog(null, "No midi device!");
			return;
		}
		record = false;
	}

	/** this saves the captured MIDI events to a MIDI file */
	public void saveMidiFile(File file) {
		// make sure there is MIDI
		if (!MIDIFound) {
			JOptionPane.showMessageDialog(null, "No midi device!");
			return;
		}
		try {
			// get the recorded sequence and write it out to the MIDI file
			int[] fileTypes = MidiSystem.getMidiFileTypes(sequence);

			if (fileTypes.length == 0) {
				JOptionPane.showMessageDialog(null, "Can't save sequence");
			} else {
				if (MidiSystem.write(sequence, fileTypes[0], file) == -1) {
					throw new IOException("Problems writing to file");
				}
			}
		} catch (Exception ex) {
			JOptionPane.showMessageDialog(null, ex);
		}
	}
}
